/* Project 8 by David Taralla - PART 2 | 18-04-2010
*  =================================================
*  This is the implementation of rvGrid.h. See this
*  header for further information on non-static
*  functions which are implemented here.
*  Some non-static functions use other functions
*  which are locale to this file only. Each precedes
*  the first function that uses them.
*/

#include <stdio.h>
#include <stdlib.h>
#include "rvPlayer.h"
#include "rvGrid.h"


/******************************************************************************
*--------------------------------- New Types ---------------------------------*
******************************************************************************/
struct rvGrid_t {
    int width;
    int height;
    int remainingCells; /* Could seem redundant with canPlay() tests;
                           however it allows to avoid the width * height
                           * 8 directions canPlay() calculation on the last
                           move that could complete the grid ! */
    
    rvGridState** content;
};


/******************************************************************************
*--------------------------------- Functions ---------------------------------*
******************************************************************************/
/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Prepare the grid for a Reversi-Othello game: it fills it with empty cells,
*  then place in middle the first four disks.
*  ----------------------------------------------------------------------------
*  Parameters :
*  =============
*  - grid: The grid to initiate
*  --------------------------------------------------------------------------*/
static void initiateGrid(rvGrid* grid) {
    // Filling grid with empty cells
    for (int row = 0; row < grid->height; row++) {
        for (int col = 0; col < grid->width; col++) {
            grid->content[row][col] = EMPTY;
        }
    }
    
    // We place the first four disks
    grid->content[grid->height / 2 - 1][grid->width / 2 - 1]
        = grid->content[grid->height / 2][grid->width / 2] = WHITE;
    grid->content[grid->height / 2 - 1][grid->width / 2]
        = grid->content[grid->height / 2][grid->width / 2 - 1] = BLACK;
    
    // Don't forget to remove the 4 cells just filled
    grid->remainingCells -= 4;
}

rvGrid* rvGridNew(const char* host, int port, int width, int height) {
    rvGrid* toReturn = malloc(sizeof(rvGrid));
    if (toReturn == NULL)
        return NULL;
    
    toReturn->width = width;
    toReturn->height = height;
    toReturn->remainingCells = width * height;
    
    toReturn->content = malloc(sizeof(rvGridState*) * height);
    if (toReturn->content == NULL)
        return NULL;
    
    // For each line, generating [width] cells
    for (int i = 0; i < height; i++) {
        toReturn->content[i] = malloc(sizeof(rvGridState) * width);
        if (toReturn->content[i] == NULL)
            return NULL;
    }
    
    initiateGrid(toReturn);
    
    // To avoid compiler warnings
    host = host;
    port = port;
    
    return toReturn;
}

void rvGridDelete(rvGrid* grid) {
    // For each line, freeing [width] cells
    for (int i = 0; i < grid->height; i++)
        free(grid->content[i]);
    
    free(grid->content);
    free(grid);
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Returns the opposite color of the parameter. This function can NEVER be used
*  with EMPTY as parameter, else it returns also WHITE.
*  ----------------------------------------------------------------------------
*  Parameters :
*  =============
*  - color: The color to inverse
*  ----------------------------------------------------------------------------
*  Return :
*  =========
*  - If [color] = WHITE : Returns BLACK
*  - If [color] = BLACK : Returns WHITE
*  - if error : Returns EMPTY
*  --------------------------------------------------------------------------*/
static rvGridState inverseColor(rvGridState color) {
    switch(color) {
        case WHITE:
            return BLACK;
        case BLACK:
            return WHITE;
        default:
            return EMPTY;
    }

    // To avoid compiler warnings
    return EMPTY;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Get the index of the first disk framing one or more disks of the opposite
*  color with the disk which has just been placed, on the left.
*  ----------------------------------------------------------------------------
*  Parameters :
*  =============
*  - grid:     The grid on which the disk is placed
*  - color:    The color of the placed disk
*  - oppColor: The opposite color of the placed disk
*  - row:      The row that contains the placed disk
*  - col:      The column that contains the placed disk
*  ----------------------------------------------------------------------------
*  Return :
*  =========
*  Index of the first framing disk (see description). If no framing disk has
*  been found, return -1.
*  --------------------------------------------------------------------------*/
static int getLeftIndexInRow(rvGrid* grid, rvGridState color,
                             rvGridState oppColor, int row, int col) {
    // If we are too near of the border, or no required color next, none found
    if (col - 1 <= 0 || grid->content[row][col - 1] != oppColor)
        return -1;
    
    for (int i = col - 2; i > -1; i--) {
        if (grid->content[row][i] == EMPTY)
            return -1;
        else if (grid->content[row][i] == color)
            return i;
        /* else if (grid->content[row][i] == oppColor)
            -> keep searching */
    }
    
    return -1;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Same work than getLeftIndexInRow, but on the right.
*  --------------------------------------------------------------------------*/
static int getRightIndexInRow(rvGrid* grid, rvGridState color,
                              rvGridState oppColor, int row, int col) {
    // If we are too near of the border, or no required color next, none found
    if (col + 2 >= grid->width || grid->content[row][col + 1] != oppColor)
        return -1;
    
    for (int i = col + 2; i < grid->width; i++) {
        if (grid->content[row][i] == EMPTY)
            return -1;
        else if (grid->content[row][i] == color)
            return i;
        /* else if (grid->content[row][i] == oppColor)
            -> keep searching */
    }
    
    return -1;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Same work than getLeftIndexInRow, but below.
*  --------------------------------------------------------------------------*/
static int getBelowIndexInCol(rvGrid* grid, rvGridState color,
                              rvGridState oppColor, int row, int col) {
    // If we are too near of the border, or no required color next, none found
    if (row + 2 >= grid->height || grid->content[row + 1][col] != oppColor)
        return -1;
    
    for (int i = row + 2; i < grid->height; i++) {
        if (grid->content[i][col] == EMPTY)
            return -1;
        else if (grid->content[i][col] == color)
            return i;
        /* else if (grid->content[i][col] == oppColor)
            -> keep searching */
    }
    
    return -1;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Same work than getLeftIndexInRow, but above.
*  --------------------------------------------------------------------------*/
static int getAboveIndexInCol(rvGrid* grid, rvGridState color,
                              rvGridState oppColor, int row, int col) {
    // If we are too near of the border, or no required color next, none found
    if (row - 1 <= 0 || grid->content[row - 1][col] != oppColor)
        return -1;
    
    for (int i = row - 2; i > -1; i--) {
        if (grid->content[i][col] == EMPTY)
            return -1;
        else if (grid->content[i][col] == color)
            return i;
        /* else if (grid->content[i][col] == oppColor)
            -> keep searching */
    }
    
    return -1;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Same work than getLeftIndexInRow, but diagonally to top-left corner.
*  --------------------------------------------------------------------------*/
static int getTLIndexInDiag(rvGrid* grid, rvGridState color,
                           rvGridState oppColor, int row, int col) {
    // If we are too near of the border, or no required color next, none found
    if (row - 1 <= 0 || col - 1 <= 0 ||
        grid->content[row - 1][col - 1] != oppColor)
        return -1;
    
    for (int i = 2; row - i > -1 && col - i > -1; i++) {
        if (grid->content[row - i][col - i] == EMPTY)
            return -1;
        else if (grid->content[row - i][col - i] == color)
            return i;
        /* else if (grid->content[row - i][col - i] == oppColor)
            -> keep searching */
    }
    
    return -1;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Same work than getLeftIndexInRow, but diagonally to bottom-right corner.
*  --------------------------------------------------------------------------*/
static int getBRIndexInDiag(rvGrid* grid, rvGridState color,
                              rvGridState oppColor, int row, int col) {
    // If we are too near of the border, or no required color next, none found
    if (row + 1 >= grid->height || col + 1 >= grid->width ||
        grid->content[row + 1][col + 1] != oppColor)
        return -1;
    
    for (int i = 2; row + i < grid->height && col + i < grid->width; i++) {
        if (grid->content[row + i][col + i] == EMPTY)
            return -1;
        else if (grid->content[row + i][col + i] == color)
            return i;
        /* else if (grid->content[row + i][col + i] == oppColor)
            -> keep searching */
    }
    
    return -1;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Same work than getLeftIndexInRow, but diagonally to top-right corner.
*  --------------------------------------------------------------------------*/
static int getTRIndexInDiag(rvGrid* grid, rvGridState color,
                           rvGridState oppColor, int row, int col) {
    // If we are too near of the border, or no required color next, none found
    if (row - 1 <= 0 || col + 1 >= grid->width ||
        grid->content[row - 1][col + 1] != oppColor)
        return -1;
    
    for (int i = 2; row - i > -1 && col + i < grid->width; i++) {
        if (grid->content[row - i][col + i] == EMPTY)
            return -1;
        else if (grid->content[row - i][col + i] == color)
            return i;
        /* else if (grid->content[row - i][col + i] == oppColor)
            -> keep searching */
    }
    
    return -1;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Same work than getLeftIndexInRow, but diagonally to bottom-left corner.
*  --------------------------------------------------------------------------*/
static int getBLIndexInDiag(rvGrid* grid, rvGridState color,
                           rvGridState oppColor, int row, int col) {
    // If we are too near of the border, or no required color next, none found
    if (row + 1 >= grid->height || col - 1 <= 0 ||
        grid->content[row + 1][col - 1] != oppColor)
        return -1;
    
    for (int i = 2; row + i < grid->height && col - i > -1; i++) {
        if (grid->content[row + i][col - i] == EMPTY)
            return -1;
        else if (grid->content[row + i][col - i] == color)
            return i;
        /* else if (grid->content[row + i][col - i] == oppColor)
            -> keep searching */
    }
    
    return -1;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Get indexes of framing disks around the one located in [row][column]. These
*  are stored in a list this order :
*  0 : index framing on the left,
*  1 : index framing on the right,
*  2 : index framing below,
*  3 : index framing above,
*  4 : index framing diagonally, oriented top-left,
*  5 : index framing diagonally, oriented bottom-right,
*  6 : index framing diagonally, oriented top-right,
*  7 : index framing diagonally, oriented bottom-left.
*  ----------------------------------------------------------------------------
*  Parameters :
*  =============
*  - list:     The list where indexes will be stored
*  - grid:     The grid on which the disk is placed
*  - color:    The color of the placed disk
*  - oppColor: The opposite color of the placed disk
*  - row:      The row that contains the placed disk
*  - col:      The column that contains the placed disk
*  --------------------------------------------------------------------------*/
static void getFramingIndexes(int* list, rvGrid* grid, rvGridState color, 
                              int row, int column) {
    rvGridState oppColor = inverseColor(color);

    list[0] = getLeftIndexInRow(grid, color, oppColor, row, column);
    list[1] = getRightIndexInRow(grid, color, oppColor, row, column);
    list[2] = getBelowIndexInCol(grid, color, oppColor, row, column);
    list[3] = getAboveIndexInCol(grid, color, oppColor, row, column);
    list[4] = getTLIndexInDiag(grid, color, oppColor, row, column);
    list[5] = getBRIndexInDiag(grid, color, oppColor, row, column);
    list[6] = getTRIndexInDiag(grid, color, oppColor, row, column);
    list[7] = getBLIndexInDiag(grid, color, oppColor, row, column);
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Counts the number of disks of one color on the grid
*  ----------------------------------------------------------------------------
*  Parameters :
*  =============
*  - grid: The grid on which we are working
*  - color: The color of the player we want to know wether he can play or not
*  ----------------------------------------------------------------------------
*  Return :
*  =========
*  - Returns true if he can play
*  - Else returns false
*  --------------------------------------------------------------------------*/
static int countDisks(rvGrid* grid, rvGridState color) {
    int count = 0;
    for (int i = 0; i < grid->height; i++) {
        for (int j = 0; j < grid->width; j++) {
            if (grid->content[i][j] == color)
                count++;
        }
    }

    return count;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Tells if a cell is empty or not.
*  ----------------------------------------------------------------------------
*  Parameters :
*  =============
*  - grid:     The grid on which the disk is placed
*  - row:      The row of the cell we check
*  - col:      The column of the cell we check
*  ----------------------------------------------------------------------------
*  Return :
*  =========
*  Index of the first framing disk (see description). If no framing disk has
*  been found, return -1.
*  --------------------------------------------------------------------------*/
static bool isCellEmpty(rvGrid* grid, int row, int col) {
    if (grid->content[row][col] != EMPTY)
        return false;

    return true;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Tells if no possible framing move is possible with the current searched
*  cell.
*  ----------------------------------------------------------------------------
*  Parameters :
*  =============
*  - list : The list which contains the indexes of the 8 possible framing with
*           the current searched cell. If the whole 8 are set to -1, no
*           possible framing.
*  ----------------------------------------------------------------------------
*  Return :
*  =========
*  - Returns true if no framing move is possible with the current searched cell
*  - Else returns false
*  --------------------------------------------------------------------------*/
static bool noneFound(int* list) {
    for (int i = 0; i < 8; i++) {
        if (list[i] != -1)
            return false;
    }

    return true;
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Tells if the last move was a winning move or a move leading to a draw, for
*  the player who made it.
*  ----------------------------------------------------------------------------
*  Parameters :
*  =============
*  - grid: The grid on which we are working
*  - currPlayerColor: The color of the current player
*  ----------------------------------------------------------------------------
*  Return :
*  =========
*  - 0 if it was a normal move (no one win, game didn't end)
*  - 1 if player of currPlayerColor color ended game and won
*  - 2 if player of currPlayerColor color ended game and lost
*  - 3 if player of currPlayerColor color ended game with a draw
*  --------------------------------------------------------------------------*/
static int getWinningMoveType(rvGrid* grid, rvGridState currPlayerColor) {
    /* If no more space available, OR next player can't play and current player
       won't also be able to play after, consider disk counts and conclude */
    if (grid->remainingCells <= 0
        || (!canPlay(grid, inverseColor(currPlayerColor))
            && !canPlay(grid, currPlayerColor))) {      
        int currColorCount = countDisks(grid, currPlayerColor);
        int nextColorCount = countDisks(grid, inverseColor(currPlayerColor));

        // If actual player have more disks than other, winning move
        if (currColorCount > nextColorCount)
            return 1;
        // Else maybe a draw
        else if (currColorCount == nextColorCount)
            return 3;
        // Else terminal move but loosing move for actual player
        return 2;
    }
    
    // Else not a winning move ; continuing game normally
    return 0;
}

int rvGridAddDisk(rvGrid* grid, rvGridState color, int row, int column) {
    if (!isCellEmpty(grid, row, column))
        return -1;
    
    int framingDiskIndexes[8];
    getFramingIndexes(framingDiskIndexes, grid, color, row, column);
    
    if (noneFound(framingDiskIndexes))
        return -1;

    grid->content[row][column] = color;
    grid->remainingCells--;

    // Replacing from the left...
    if (framingDiskIndexes[0] != -1)
        for (int i = framingDiskIndexes[0] + 1; i < column; i++)
            grid->content[row][i] = color;
    
    // Replacing from the right...
    if (framingDiskIndexes[1] != -1)
        for (int i = framingDiskIndexes[1] - 1; i > column; i--)
            grid->content[row][i] = color;
    
    // Replacing from below...
    if (framingDiskIndexes[2] != -1)
        for (int i = framingDiskIndexes[2] - 1; i > row; i--)
            grid->content[i][column] = color;
    
    // Replacing from above...
    if (framingDiskIndexes[3] != -1)
        for (int i = framingDiskIndexes[3] + 1; i < row; i++)
            grid->content[i][column] = color;
    
    // Replacing diagonnaly from top-left...
    if (framingDiskIndexes[4] != -1)
        for (int i = framingDiskIndexes[4] - 1; i > 0; i--)
            grid->content[row - i][column - i] = color;
    
    // Replacing diagonnaly from bottom-right...
    if (framingDiskIndexes[5] != -1)
        for (int i = framingDiskIndexes[5] - 1; i > 0; i--)
            grid->content[row + i][column + i] = color;
    
    // Replacing diagonnaly from top-right...
    if (framingDiskIndexes[6] != -1)
        for (int i = framingDiskIndexes[6] - 1; i > 0; i--)
            grid->content[row - i][column + i] = color;

    // Replacing diagonnaly from bottom-left...
    if (framingDiskIndexes[7] != -1)
        for (int i = framingDiskIndexes[7] - 1; i > 0; i--)
            grid->content[row + i][column - i] = color;
    
    return getWinningMoveType(grid, color);
}

/* ----------------------------------------------------------------------------
*  Description :
*  ==============
*  Converts the rvGridState contained at the position [row], [col] into a
*  character usable by the rvGridDisplay function.
*  ----------------------------------------------------------------------------
*  Parameters :
*  =============
*  - grid: The grid on which we are working
*  - row:  The row of the rvGridState to convert
*  - col:  The column of the rvGridState to convert
*  ----------------------------------------------------------------------------
*  Return :
*  =========
*  - If EMPTY, returns   (simple space)
*  - If WHITE, returns X
*  - If BLACK, returns O
*  --------------------------------------------------------------------------*/
static char colorToChar(rvGrid* grid, int row, int col) {
    switch(grid->content[row][col]) {
        case EMPTY:
            return ' ';
        case WHITE:
            return 'X';
        case BLACK:
            return 'O';
    }

    // To avoid compiler warnings
    return 'O';
}

void rvGridDisplay(rvGrid* grid, FILE* output) {
    fprintf(output, "It's your turn! Actual state of the grid:\n");
    for (int row = 0; row < grid->height; row++) {
        for (int col = 0; col < grid->width; col++)
            fprintf(output, "%c ", colorToChar(grid, row, col));
        
        fprintf(output, "\n");
    }
}

bool canPlay(rvGrid* grid, rvGridState color) {
    int indexesList[grid->width * grid->height][8];
    int cell = 0;

    for (int i = 0; i < grid->width; i++) {
        for (int j = 0; j < grid->height; j++) {
            if (isCellEmpty(grid, i, j)) {
                getFramingIndexes(indexesList[cell], grid, color, i, j);
                if (!noneFound(indexesList[cell]))
                    return true;
            }
            cell++;
        }
    }

    return false;
}
